"""Script to train model."""
import argparse
from datetime import datetime
import os
import time

import torch
import torch.utils.tensorboard
import wandb
import yoco

from torchinfo import summary

import sdfest

from sdfest.vae import utils
from sdfest.vae import sdf_utils
from sdfest.vae.sdf_vae import SDFVAE
from sdfest.vae.sdf_dataset import SDFDataset

from sdfest.differentiable_renderer import Camera, render_depth_gpu

from sdfest.initialization import pointset_utils

if "DISPLAY" not in os.environ or "localhost" in os.environ["DISPLAY"]:
    print("Using EGL instead of pyglet for OffscreenRendering")
    os.environ["PYOPENGL_PLATFORM"] = "egl"


def pc_loss(points, position, orientation, scale, sdf):
    """Compute trilinerly interpolated SDF value at the points positions.

    Args:
        points:
            Mx4 pointcloud in camera frame.
        position:
            Position of SDF center in camera frame.
        orientation:
            Quaternion representing orientation of SDF.
        scale:
            Half-width of SDF volume.
        sdf:
            Volumetric signed distance field.
    Returns:
        Trilinearly interpolated distance at the passed points
        0 if outside of SDF volume.
    """
    q = orientation / torch.norm(orientation)  # to get normalization gradients
    obj_points = points - position.unsqueeze(0)

    # Quaternion to rotation matrix
    # Note that we use conjugate here since we want to transform points from
    # world to object coordinates and the quaternion describes rotation of
    # object coordinate system in world coordinates.
    R = obj_points.new_zeros(3, 3)

    R[0, 0] = 1 - 2 * (q[1] * q[1] + q[2] * q[2])
    R[0, 1] = 2 * (q[0] * q[1] + q[2] * q[3])
    R[0, 2] = 2 * (q[0] * q[2] - q[3] * q[1])

    R[1, 0] = 2 * (q[0] * q[1] - q[2] * q[3])
    R[1, 1] = 1 - 2 * (q[0] * q[0] + q[2] * q[2])
    R[1, 2] = 2 * (q[1] * q[2] + q[3] * q[0])

    R[2, 0] = 2 * (q[0] * q[2] + q[3] * q[1])
    R[2, 1] = 2 * (q[1] * q[2] - q[3] * q[0])
    R[2, 2] = 1 - 2 * (q[0] * q[0] + q[1] * q[1])

    obj_points = (R @ obj_points.T).T

    # Transform to canonical coordintes obj_point in [-1,1]^3
    obj_point = obj_points / scale

    # Compute cell and cell position
    res = sdf.shape[0]  # assuming same resolution along each axis
    grid_size = 2.0 / (res - 1)
    c = torch.floor((obj_point + 1.0) * (res - 1) * 0.5)
    mask = torch.logical_or(
        torch.min(c, dim=1)[0] < 0, torch.max(c, dim=1)[0] > res - 2
    )
    c = torch.clip(c, 0, res - 2)  # base cell index of each point
    cell_position = c * grid_size - 1.0  # base cell position of each point
    sdf_indices = c.new_empty((obj_point.shape[0], 8), dtype=torch.long)
    sdf_indices[:, 0] = c[:, 0] * res**2 + c[:, 1] * res + c[:, 2]
    sdf_indices[:, 1] = c[:, 0] * res**2 + c[:, 1] * res + c[:, 2] + 1
    sdf_indices[:, 2] = c[:, 0] * res**2 + (c[:, 1] + 1) * res + c[:, 2]
    sdf_indices[:, 3] = c[:, 0] * res**2 + (c[:, 1] + 1) * res + c[:, 2] + 1
    sdf_indices[:, 4] = (c[:, 0] + 1) * res**2 + c[:, 1] * res + c[:, 2]
    sdf_indices[:, 5] = (c[:, 0] + 1) * res**2 + c[:, 1] * res + c[:, 2] + 1
    sdf_indices[:, 6] = (c[:, 0] + 1) * res**2 + (c[:, 1] + 1) * res + c[:, 2]
    sdf_indices[:, 7] = (c[:, 0] + 1) * res**2 + (c[:, 1] + 1) * res + c[:, 2] + 1
    sdf_view = sdf.view([-1])
    point_cell_position = (obj_point - cell_position) / grid_size  # [0,1]^3
    sdf_values = torch.take(sdf_view, sdf_indices)

    # trilinear interpolation
    sdf_value = sdf_values.new_empty(obj_points.shape[0])
    # sdf_value = obj_point[:, 0]
    sdf_value = (
        (
            sdf_values[:, 0] * (1 - point_cell_position[:, 0])
            + sdf_values[:, 4] * point_cell_position[:, 0]
        )
        * (1 - point_cell_position[:, 1])
        + (
            sdf_values[:, 2] * (1 - point_cell_position[:, 0])
            + sdf_values[:, 6] * point_cell_position[:, 0]
        )
        * point_cell_position[:, 1]
    ) * (1 - point_cell_position[:, 2]) + (
        (
            sdf_values[:, 1] * (1 - point_cell_position[:, 0])
            + sdf_values[:, 5] * point_cell_position[:, 0]
        )
        * (1 - point_cell_position[:, 1])
        + (
            sdf_values[:, 3] * (1 - point_cell_position[:, 0])
            + sdf_values[:, 7] * point_cell_position[:, 0]
        )
        * point_cell_position[:, 1]
    ) * point_cell_position[
        :, 2
    ]
    sdf_value[mask] = 0
    return sdf_value


def train(config):
    """Train the model.

    Args:
        config: The training and model config.
    """
    batch_size = wandb.config.batch_size = config["batch_size"]
    iterations = wandb.config.iterations = config["iterations"]
    learning_rate = wandb.config.learning_rate = config["learning_rate"]
    latent_size = wandb.config.latent_size = config["latent_size"]
    l2_large_weight = wandb.config.l2_large_weight = config["l2_large_weight"]
    l2_small_weight = wandb.config.l2_small_weight = config["l2_small_weight"]
    l1_large_weight = wandb.config.l1_large_weight = config["l1_large_weight"]
    l1_small_weight = wandb.config.l1_small_weight = config["l1_small_weight"]
    kld_weight = wandb.config.kld_weight = config["kld_weight"]
    pc_weight = wandb.config.pc_weight = config["pc_weight"]
    encoder_dict = wandb.config.encoder_dict = config["encoder"]
    decoder_dict = wandb.config.decoder_dict = config["decoder"]
    dataset_path = wandb.config.dataset_path = config["dataset_path"]
    tsdf = wandb.config.tsdf = config["tsdf"]

    # init dataset, dataloader, model and optimizer
    dataset = SDFDataset(dataset_path)
    data_loader = torch.utils.data.DataLoader(
        dataset=dataset, batch_size=batch_size, shuffle=True, drop_last=True
    )

    camera = Camera(640, 480, 320, 320, 320, 240, pixel_center=0.5)

    if "device" not in config or config["device"] is None:
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    else:
        device = torch.device(config["device"])

    sdfvae = SDFVAE(
        sdf_size=64,
        latent_size=latent_size,
        encoder_dict=encoder_dict,
        decoder_dict=decoder_dict,
        device=device,
        tsdf=tsdf,
    ).to(device)

    print(str(sdfvae))

    summary(sdfvae, (1, 1, 64, 64, 64), device=device)

    optimizer = torch.optim.Adam(sdfvae.parameters(), lr=learning_rate)

    # load checkpoint if provided
    if "checkpoint" in config and config["checkpoint"] is not None:
        # TODO: checkpoint should always go together with model config!
        model, optimizer, current_iteration, run_name, epoch = utils.load_checkpoint(
            config["checkpoint"], sdfvae, optimizer
        )
    else:
        current_iteration = 0
        current_epoch = 0
        run_name = f"sdfvae_{datetime.now().strftime('%Y-%m-%d_%H-%M-%S-%f')}"

    # create SummaryWriter for logging and intermediate output
    writer = torch.utils.tensorboard.SummaryWriter(f"runs/" + run_name)

    model_base_path = os.path.join(os.getcwd(), "models", run_name)
    program_starts = time.time()
    warm_up_iterations = 1000
    stop = False
    while current_iteration <= iterations:
        current_epoch += 1
        for sdf_volumes in data_loader:
            sdf_volumes = sdf_volumes.to(device)

            # train first N iters with SDF instead of TSDF to stabilize early training
            if current_iteration > warm_up_iterations:
                sdfvae.prepare_input(sdf_volumes)

            recon_sdf_volumes, mean, log_var, z = sdfvae(
                sdf_volumes, enforce_tsdf=False
            )

            if tsdf is not False and current_iteration > warm_up_iterations:
                # during training, clamp the reconstructed SDF only where both target
                # and output are outside the TSDF range
                mask = torch.logical_and(
                    torch.abs(sdf_volumes) >= tsdf, torch.abs(recon_sdf_volumes) >= tsdf
                )
                recon_sdf_volumes_temp = recon_sdf_volumes
                recon_sdf_volumes = recon_sdf_volumes_temp.clone()
                recon_sdf_volumes[mask] = recon_sdf_volumes_temp[mask].clamp(
                    -tsdf, tsdf
                )

            # Compute losses, use negative log-likelihood here
            # Note: for average negative log-likelihood all of these have to be divided
            # by the batch size. Probably would be better to keep losses comparable for
            # varying batch size.
            l1_error = torch.abs(recon_sdf_volumes - sdf_volumes)
            l2_error = l1_error**2
            loss_l2_small = torch.sum(l2_error[torch.abs(sdf_volumes) < 0.1])
            loss_l2_large = torch.sum(l2_error[torch.abs(sdf_volumes) >= 0.1])
            loss_l1_small = torch.sum(l1_error[torch.abs(sdf_volumes) < 0.1])
            loss_l1_large = torch.sum(l1_error[torch.abs(sdf_volumes) >= 0.1])
            loss_pc = 0

            depth_images = torch.empty((batch_size, 480, 640), device=device)
            pointclouds = []
            # compute point clouds from sdf volumes at zero crossings
            for recon_sdf_volume, sdf_volume, depth_image in zip(
                recon_sdf_volumes, sdf_volumes, depth_images
            ):
                with torch.no_grad():
                    import random
                    import math

                    u1, u2, u3 = random.random(), random.random(), random.random()
                    q = (
                        torch.tensor(
                            [
                                math.sqrt(1 - u1) * math.sin(2 * math.pi * u2),
                                math.sqrt(1 - u1) * math.cos(2 * math.pi * u2),
                                math.sqrt(u1) * math.sin(2 * math.pi * u3),
                                math.sqrt(u1) * math.cos(2 * math.pi * u3),
                            ]
                        )
                        .unsqueeze(0)
                        .to(device)
                    )
                    s = torch.Tensor([1.0]).to(device)
                    p = torch.Tensor([0.0, 0.0, -5.0]).to(device)

                    depth_image = render_depth_gpu(
                        sdf_volume[0],
                        p,
                        q,
                        s,
                        threshold=0.01,
                        camera=camera,
                    )
                    pointcloud = pointset_utils.depth_to_pointcloud(depth_image, camera)
                loss_pc = loss_pc + torch.sum(
                    pc_loss(pointcloud, p, q[0], s, recon_sdf_volume[0]) ** 2
                )

            loss_kld = -0.5 * torch.sum(1 + log_var - mean.pow(2) - log_var.exp())

            loss = (
                l2_small_weight * loss_l2_small
                + l2_large_weight * loss_l2_large
                + l1_small_weight * loss_l1_small
                + l1_large_weight * loss_l1_large
                + pc_weight * loss_pc
                + loss_kld
                * (kld_weight if current_iteration > warm_up_iterations else 0)
            )

            print(f"Iteration {current_iteration}, epoch {current_epoch}, loss {loss}")

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            writer.add_scalar("loss", loss.item(), current_iteration)
            writer.add_scalar("loss l2 small", loss_l2_small.item(), current_iteration)
            writer.add_scalar("loss l2 large", loss_l2_large.item(), current_iteration)
            writer.add_scalar("loss l1 small", loss_l1_small.item(), current_iteration)
            writer.add_scalar("loss l1 large", loss_l1_large.item(), current_iteration)
            writer.add_scalar("loss pc", loss_pc.item(), current_iteration)
            writer.add_scalar("loss kl", loss_kld, current_iteration)

            wandb.log(
                {
                    "total loss": loss.item(),
                    "loss l2 small": loss_l2_small.item(),
                    "loss l2 large": loss_l2_large.item(),
                    "loss l1 small": loss_l1_small.item(),
                    "loss l1 large": loss_l1_large.item(),
                    "loss kl": loss_kld.item(),
                    "loss pc": loss_pc.item(),
                },
                step=current_iteration,
            )

            current_iteration += 1

            with torch.no_grad():
                if current_iteration % 1000 == 0:
                    # show one reconstruction
                    mean = mean[0].cpu()
                    figure = sdf_utils.visualize_sdf_reconstruction(
                        sdf_volumes[0, 0].cpu().numpy(),
                        recon_sdf_volumes[0, 0].cpu().numpy(),
                    )
                    if figure.gca().has_data():
                        writer.add_figure(
                            tag="reconstruction",
                            figure=figure,
                            global_step=current_iteration,
                        )
                        wandb.log({"reconstruction": figure}, step=current_iteration)

                    # generate 8 samples from the prior
                    output, _ = sdfvae.inference(n=8, enforce_tsdf=True)
                    figure = sdf_utils.visualize_sdf_batch(
                        output.squeeze().cpu().numpy()
                    )
                    if figure.gca().has_data():
                        writer.add_figure(
                            tag="samples from prior",
                            figure=figure,
                            global_step=current_iteration,
                        )
                        wandb.log(
                            {"samples from prior": figure}, step=current_iteration
                        )

                    # generate 4 samples from prior
                    output, _ = sdfvae.inference(n=4, enforce_tsdf=True)
                    figure = sdf_utils.visualize_sdf_batch_columns(
                        output.squeeze().cpu().numpy()
                    )
                    if figure.gca().has_data():
                        writer.add_figure(
                            tag="samples from prior, columns",
                            figure=figure,
                            global_step=current_iteration,
                        )
                        wandb.log(
                            {"samples from prior, columns": figure},
                            step=current_iteration,
                        )

            #     if current_iteration % 10000 == 0:
            #         os.makedirs(model_base_path, exist_ok=True)
            #         checkpoint_path = os.path.join(model_base_path, F"{current_iteration}.ckp")
            #         utils.save_checkpoint(
            #             path=checkpoint_path,
            #             model=sdfvae,
            #             optimizer=optimizer,
            #             iteration=current_iteration,
            #             run_name=run_name)

            if current_iteration > iterations:
                break
        if current_iteration > iterations:
            break

    now = time.time()
    print("It has been {0} seconds since the loop started".format(now - program_starts))

    # save the final model
    torch.save(sdfvae.state_dict(), os.path.join(wandb.run.dir, f"{wandb.run.name}.pt"))
    config_path = os.path.join(wandb.run.dir, f"{wandb.run.name}.yaml")
    config["model"] = os.path.join(".", f"{wandb.run.name}.pt")
    yoco.save_config_to_file(config_path, config)


def main():
    """Entry point of the program."""
    # define the arguments
    parser = argparse.ArgumentParser(description="Training script for SDFVAE.")

    # parse arguments
    parser.add_argument("--checkpoint")
    parser.add_argument("--dataset_path", required=True)
    parser.add_argument("--batch_size", type=lambda x: int(float(x)))
    parser.add_argument("--iterations", type=lambda x: int(float(x)))
    parser.add_argument("--latent_size", type=lambda x: int(float(x)))
    parser.add_argument("--tsdf", type=utils.str_to_tsdf, default=False)
    parser.add_argument("--kld_weight", type=lambda x: float(x))
    parser.add_argument("--l2_large_weight", type=lambda x: float(x))
    parser.add_argument("--l2_small_weight", type=lambda x: float(x))
    parser.add_argument("--l1_large_weight", type=lambda x: float(x))
    parser.add_argument("--l1_small_weight", type=lambda x: float(x))
    parser.add_argument("--pc_weight", type=lambda x: float(x))
    parser.add_argument("--sign_weight", type=lambda x: float(x))
    parser.add_argument("--bce_weight", type=lambda x: float(x))
    parser.add_argument("--learning_rate", type=lambda x: float(x))
    parser.add_argument("--device")
    parser.add_argument("--config", default="configs/default.yaml", nargs="+")
    config = yoco.load_config_from_args(
        parser, search_paths=[".", "~/.sdfest/", sdfest.__path__[0]]
    )
    train(config)


if __name__ == "__main__":
    wandb.init(project="sdf-vae")
    main()



"""Common utility functions."""
import os
from typing import Optional

from cpas_toolbox.utils import download
import torch
import yoco


def load_model_weights(
    path: str,
    model: torch.nn.Module,
    map_location: torch.device,
    model_weights_url: Optional[str] = None,
) -> None:
    """Load model weights from path or download weights from URL if file does not exist.

    Args:
        path: Path to model weights.
        model: Path to model weights.
        map_location: See torch.load.
        model_weights_url: URL to download model weights from if path does not exist.
    """
    resolved_path = yoco.resolve_path(
        path, search_paths=[".", "~/.sdfest/model_weights/"]
    )

    if not os.path.exists(resolved_path):
        if model_weights_url is not None:
            if not os.path.isabs(resolved_path):
                resolved_path = os.path.expanduser(
                    os.path.join("~/.sdfest/model_weights", resolved_path)
                )
            os.makedirs(os.path.dirname(resolved_path), exist_ok=True)
            print(f"Model weights {resolved_path} not found.")
            print(f"Downloading from {model_weights_url}")
            download(model_weights_url, resolved_path)
        else:
            print(f"Model weights {resolved_path} not found. Aborting.")
            exit(0)

    state_dict = torch.load(resolved_path, map_location=map_location)
    model.load_state_dict(state_dict)



import setuptools
import glob
import os

with open("PYPIREADME.md", "r") as fh:
    long_description = fh.read()

config_files = [
    os.path.relpath(path, "sdfest")
    for path in glob.glob("sdfest/**/configs/**/*.yaml", recursive=True)
]

setuptools.setup(
    name="sdfest",
    version="0.1.1",
    author="Leonard Bruns",
    author_email="roym899@gmail.com",
    description="6-DoF pose, scale, and shape estimation architecture",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/roym899/sdfest",
    packages=setuptools.find_packages(),
    package_data={
        "sdfest": [
            "differentiable_renderer/csrc/sdf_renderer.cpp",
            "differentiable_renderer/csrc/sdf_renderer_cuda.cu",
        ]
        + config_files
    },
    install_requires=[
        "cpas_toolbox",
        "ffmpeg-python",
        "healpy",
        "joblib",
        "matplotlib",
        # "mesh-to-sdf, removed due to https://github.com/marian42/mesh_to_sdf/issues/38
        "ninja",
        "numpy",
        "open3d",
        "pandas",
        "pynput",
        "pyrender",
        "PySide2",
        "tensorboard",
        "trimesh",
        "scipy",
        "scikit-image",
        "tabulate",
        "tqdm",
        "torch",
        "torchinfo",
        "torchvision",
        "wandb",
        "yoco",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
    ],
    python_requires=">=3.7",
)
# include_package_data=True,
# ext_modules=[
#     CUDAExtension(
#         "sdf_renderer_cuda",
#         [
#             "sdfest/differentiable_renderer/csrc/sdf_renderer.cpp",
#             "sdfest/differentiable_renderer/csrc/sdf_renderer_cuda.cu",
#         ],
#     )
# ],
# cmdclass={"build_ext": BuildExtension},



"""Script to run randomized rendering evaluation on synthetic data.

Usage:
  python rendering_evaluation.py --config configs/config.yaml --data_path ./data/
  --out_folder ./results

See configs/rendering_evaluation.yaml for all supported arguments.
See simple_setup for pipeline parameters.

Specific parameters:
    log_folder:
        if passed each optimization step is logged and can be played back with
        play_log.py
    visualize_optimization: whether to visualize optimization while at it
    visualize_points: whether to show result pointclouds after optimization
    visualize_meshes: whether to show result mesh after optimization
    camera_distance: mesh distance from the camera
    num_views: list of number of views to evaluate for each mesh
    mesh_scale: the applied scale, see rel_scale
    rel_scale:
        if True, the original mesh will be scaled by mesh_scale, if False the original
        mesh will be scaled such that its largest extent is mesh_scale * 2
    samples: number of evaluation samples
    ablation_configs:
        used to specify specific settings for ablation study
        dictionary, in which each key maps to a config dictionary which will be applied
        on existing settings
    metrics:
        dictionary of metrics to evaluate
        each key, is interpreted as the name of the metric, each value has to be a dict
        containing f and kwargs, where f is fully qualified name of the function to
        evaluate and kwargs is a dictionary of keyword arguments if applicable
        f gets ground truth points as first, and estimated points as second
        parameter
    seed:
        seed used for view generation and sampling of points
"""

import argparse
from collections import defaultdict
import copy
from datetime import datetime
import glob
import math
import os
from pydoc import locate
import random
import time
from typing import Dict, List, Optional

import numpy as np
import open3d as o3d
import torch
from tqdm import tqdm
import yoco

import sdfest
from sdfest.initialization import quaternion_utils
from sdfest.estimation import synthetic
from sdfest.differentiable_renderer import Camera
from sdfest.estimation.simple_setup import SDFPipeline


def glob_exts(path: str, exts: List[str]) -> List[str]:
    """Return all files in a nested directory with extensions matching.

    Directory is scanned recursively.

    Args:
        path: root path to search
        exts: extensions that will be checked, must include separator (e.g., ".obj")

    Returns:
        List of paths in the directory with matching extension.
    """
    files = []
    for ext in exts:
        files.extend(glob.glob(os.path.join(path, f"**/*{ext}"), recursive=True))

    return files


class Evaluator:
    """Class to evaluate SDF pipeline on synthetic data."""

    def __init__(self, config: dict) -> None:
        """Construct evaluator and initialize pipeline."""
        self.base_config = config
        self.cam = Camera(**self.base_config["camera"])

    def run(self) -> None:
        """Run the evaluation."""
        o3d.utility.set_verbosity_level(o3d.utility.VerbosityLevel.Warning)
        if self.base_config["ablation_configs"]:
            ablation_results_dict = {}
            for name, ablation_config in self.base_config["ablation_configs"].items():
                config = yoco.load_config(
                    ablation_config, copy.deepcopy(self.base_config)
                )
                self._set_seed(config["seed"])
                ablation_results_dict[name] = self._evaluate_config(config)
            self._save_and_print_results(ablation_results_dict)
        else:
            self._set_seed(self.base_config["seed"])
            results_dict = self._evaluate_config(self.base_config)
            self._save_and_print_results(results_dict)

    @staticmethod
    def _set_seed(seed: int = 0) -> None:
        random.seed(seed)

    def _evaluate_config(self, config: dict) -> dict:
        results_dict = {}
        self.pipeline = SDFPipeline(config)
        for views in config["num_views"]:
            metrics_list = []
            files = glob_exts(config["data_path"], [".obj", ".off"])
            files.sort()
            for file in tqdm(files):
                metrics = self._evaluate_file(file, views, config)
                metrics_list.append(metrics)

            results_dict[views] = Evaluator._compute_metric_statistics(metrics_list)
        return results_dict

    def _save_and_print_results(self, results_dict: Dict) -> None:
        """Save results and config to yaml file and print results as table.

        Args:
            results_dict:
                dictionary containing the results that should be saved
        """
        os.makedirs(self.base_config["out_folder"], exist_ok=True)
        run_name = self.base_config["run_name"]
        filename = (
            f"rend_eval_{run_name}_{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}.yaml"
        )
        out_path = os.path.join(self.base_config["out_folder"], filename)
        combined_dict = {**self.base_config, "results": results_dict}
        yoco.save_config_to_file(out_path, combined_dict)
        print(f"Results saved to: {out_path}")

    @staticmethod
    def _compute_metric_statistics(metrics_list: List) -> Dict:
        """Compute mean and standard deviation for each metric.

        Args:
            metrics_list: metric dictionaries as returned by _evaluate_file.

        Returns:
            Statistic for each metric in the provided metrics dictionaries.
            The returned dictionary keys will be the name of the metrics. Each value
            will be another dictionary containing the keys mean, var, and std.
        """
        metric_stats = defaultdict(lambda: {"mean": 0, "var": 0})
        for metrics in metrics_list:
            for name, val in metrics.items():
                metric_stats[name]["mean"] += val

        for _, stats in metric_stats.items():
            stats["mean"] /= len(metrics_list)

        for metrics in metrics_list:
            for name, val in metrics.items():
                metric_stats[name]["var"] += (val - metric_stats[name]["mean"]) ** 2

        for _, stats in metric_stats.items():
            stats["var"] /= len(metrics_list)
            stats["std"] = math.sqrt(stats["var"])

        metric_stats = dict(metric_stats)
        return metric_stats

    def _generate_views(self, mesh: synthetic.Mesh, num_views: int) -> Dict:
        """Generate random views around mesh.

        Args:
            mesh:
                mesh to generate views of,
                position and orientation will be assumed to be in world coordinates
            num_views: number of views to generate

        Returns:
            Dictionary containing the following keys.

            depth_images:
                the depth map containing the distance along the camera's z-axis,
                shape (num_views, H, W)
            masks:
                binary mask of the object to estimate, same shape as depth_images
            color_images:
                color images of objects to estimate, shape (num_views, H, W, 3)
                note that this is currently just zero
            camera_positions:
                position of camera in world coordinates for each image,
                shape (num_views, 3)
            camera_orientations:
                orientation of camera in world-frame as normalized quaternion,
                quaternion is in scalar-last convention,
                this is the quaternion that transforms a point from camera to world,
                shape (num_views, 4)
        """
        views_dict = defaultdict(lambda: list())

        mesh.position = np.array([0.0, 0.0, 0.0], dtype=np.float32)
        mesh_position = torch.tensor(mesh.position)
        mesh_orientation = torch.tensor(mesh.orientation)

        for _ in range(num_views):
            while True:
                # OpenGL convention camera
                camera_orientation = generate_uniform_quaternion()  # ogl to world
                camera_position = mesh_position - quaternion_utils.quaternion_apply(
                    camera_orientation,
                    torch.tensor([0, 0, -self.base_config["camera_distance"]]),
                )  # transform camera position s.t. object lies on principal axis

                # Transform mesh into camera frame, now with Open3D convention camera
                camera_orientation_o3d = quaternion_utils.quaternion_multiply(
                    camera_orientation,  # ogl to world
                    torch.tensor([1.0, 0, 0, 0]),  # o3d to ogl
                )  # quaternion: o3d to world
                mesh_orientation_cam = quaternion_utils.quaternion_multiply(
                    quaternion_utils.quaternion_invert(
                        camera_orientation_o3d
                    ),  # world to o3d
                    mesh_orientation,  # obj to world
                )  # quaternion: obj to o3d
                mesh.position = np.array([0, 0, self.base_config["camera_distance"]])
                mesh.orientation = mesh_orientation_cam.numpy()
                depth_np = synthetic.draw_depth_geometry(mesh, self.cam)
                depth = torch.tensor(depth_np)

                if (depth != 0).any():
                    views_dict["depth_images"].append(depth)
                    views_dict["masks"].append(depth != 0)
                    views_dict["color_images"].append(torch.zeros(depth.shape + (3,)))
                    views_dict["camera_positions"].append(camera_position)
                    views_dict["camera_orientations"].append(camera_orientation)
                    break

                print("Warning: invalid depth generated, skipping this sample")

        mesh.position = mesh_position.numpy()
        mesh.orientation = mesh_orientation.numpy()

        return {
            k: torch.stack(v).to(self.base_config["device"])
            for k, v in views_dict.items()
        }

    def _evaluate_file(self, path: str, num_views: int, config: dict) -> dict:
        """Evaluate a single mesh.

        This will generate depth images from a few views with the mesh centered, at
        fixed distance,  to the camera.

        Args:
            path: The path of the obj file.
            num_views: The number of views to generate and use in the optimization.
            config: Configuration dictionary.

        Returns:
            Evaluation metrics as specified in config.
        """
        gt_mesh = synthetic.Mesh(
            path=path,
            scale=self.base_config["mesh_scale"],
            rel_scale=self.base_config["rel_scale"],
            center=True,
        )
        inputs = self._generate_views(gt_mesh, num_views)

        log_path = self._get_log_path()

        position, orientation, scale, shape = self.pipeline(
            **inputs,
            visualize=self.base_config["visualize_optimization"],
            log_path=log_path,
            shape_optimization=config["shape_optimization"],
        )
        out_mesh = self.pipeline.generate_mesh(shape, scale, True)

        # Output and ground truth are in world frame
        out_mesh.position = position[0].detach().cpu().numpy()
        out_mesh.orientation = orientation[0].detach().cpu().numpy()
        gt_mesh = gt_mesh.get_transformed_o3d_geometry()
        out_mesh = out_mesh.get_transformed_o3d_geometry()
        gt_mesh.paint_uniform_color([0.7, 0.4, 0.2])
        out_mesh.paint_uniform_color([0.2, 0.4, 0.7])
        gt_pts = gt_mesh.sample_points_uniformly(
            number_of_points=self.base_config["samples"], seed=self.base_config["seed"]
        )
        out_pts = out_mesh.sample_points_uniformly(
            number_of_points=self.base_config["samples"], seed=self.base_config["seed"]
        )
        gt_pts_np = np.asarray(gt_pts.points)
        out_pts_np = np.asarray(out_pts.points)

        metric_dict = {}
        for metric_name, m in self.base_config["metrics"].items():
            metric_dict[metric_name] = float(
                locate(m["f"])(gt_pts_np, out_pts_np, **m["kwargs"])
            )

        self.visualize_result(gt_mesh, out_mesh, gt_pts, out_pts, inputs)

        return metric_dict

    def visualize_result(
        self,
        mesh_1: o3d.geometry.TriangleMesh,
        mesh_2: o3d.geometry.TriangleMesh,
        pts_1: o3d.geometry.PointCloud,
        pts_2: o3d.geometry.PointCloud,
        inputs: Optional[dict] = None,
    ) -> None:
        """Visualize result of a single evaluation.

        Args:
            mesh_1: the first mesh to visualize
            mesh_2: the second mesh to visualize
            pts_1: the first pointcloud to visualize
            pts_2: the second pointcloud to visualize
            inputs: the input dictionary as producted by Evaluator._generate_view
        """
        # generate coordinate frames of cameras
        cam_meshes = []
        if inputs is not None:
            # visualize OpenGL convention camera
            for t_c2w, quat_c2w in zip(
                inputs["camera_positions"], inputs["camera_orientations"]
            ):
                frame_mesh = synthetic.Mesh(
                    mesh=o3d.geometry.TriangleMesh.create_coordinate_frame(
                        size=0.1, origin=[0, 0, 0]
                    ),
                    rel_scale=True,
                )
                frame_mesh.position = t_c2w.cpu().numpy()
                frame_mesh.orientation = quat_c2w.cpu().numpy()
                cam_meshes.append(frame_mesh.get_transformed_o3d_geometry())

        if self.base_config["visualize_meshes"]:
            # Visualize result
            o3d.visualization.draw_geometries(
                [
                    mesh_1,
                    mesh_2,
                ]
                + cam_meshes
            )
            time.sleep(0.1)

        if self.base_config["visualize_points"]:
            o3d.visualization.draw_geometries([pts_1, pts_2] + cam_meshes)
            time.sleep(0.1)

    def _get_log_path(self) -> Optional[str]:
        """Return unique filename in log folder.

        If log path is None, None will be returned.
        """
        log_path = None
        if self.base_config["log_folder"] is not None:
            os.makedirs(self.base_config["log_folder"], exist_ok=True)
            filename = f"log_{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}.pkl"
            log_path = os.path.join(self.base_config["log_folder"], filename)
        return log_path


def generate_uniform_quaternion() -> torch.Tensor:
    """Generate a normalized uniform quaternion.

    Following the method from K. Shoemake, Uniform Random Rotations, 1992.

    See: http://planning.cs.uiuc.edu/node198.html

    Returns:
        Uniformly distributed unit quaternion on the estimator's device.
    """
    u1, u2, u3 = random.random(), random.random(), random.random()
    return torch.tensor(
        [
            math.sqrt(1 - u1) * math.sin(2 * math.pi * u2),
            math.sqrt(1 - u1) * math.cos(2 * math.pi * u2),
            math.sqrt(u1) * math.sin(2 * math.pi * u3),
            math.sqrt(u1) * math.cos(2 * math.pi * u3),
        ]
    )


def main() -> None:
    """Entry point of the evaluation program."""
    parser = argparse.ArgumentParser(
        description="SDF shape and pose estimation evaluation on synthetic data"
    )
    parser.add_argument(
        "--config", default="configs/rendering_evaluation.yaml", nargs="+"
    )
    parser.add_argument("--data_path", required=True)
    parser.add_argument("--out_folder", required=True)
    config = yoco.load_config_from_args(
        parser, search_paths=[".", "~/.sdfest/", sdfest.__path__[0]]
    )

    evaluator = Evaluator(config)
    evaluator.run()


if __name__ == "__main__":
    main()



"""Module providing dataset class for NOCS datasets (CAMERA / REAL)."""
import datetime
import imghdr
import json
from glob import glob
import os
import pickle
from shutil import copyfile
import time
from typing import TypedDict, Optional

from joblib import Parallel, delayed
from scipy.spatial.transform import Rotation
import numpy as np
import open3d as o3d
import pandas as pd
import torch
from PIL import Image
from sdfest.differentiable_renderer import Camera
import yoco
from tqdm import tqdm

from sdfest.initialization import pointset_utils, quaternion_utils, so3grid
from sdfest.initialization.datasets import nocs_utils


class NOCSDataset(torch.utils.data.Dataset):
    """Dataset class for NOCS dataset.

    CAMERA* and REAL* are training sets.
    CAMERA25 and REAL275 are test data.
    Some papers use CAMERA25 as validation when benchmarking REAL275.

    Datasets can be found here:
    https://github.com/hughw19/NOCS_CVPR2019/tree/master

    Expected directory format:
        {root_dir}/real_train/...
        {root_dir}/real_test/...
        {root_dir}/gts/...
        {root_dir}/obj_models/...
        {root_dir}/camera_composed_depth/...
        {root_dir}/val/...
        {root_dir}/train/...
    Which is easily obtained by downloading all the provided files and extracting them
    into the same directory.

    Necessary preprocessing of this data is performed during first initialization per
    and is saved to
        {root_dir}/sdfest_pre/...
    """

    num_categories = 7
    category_id_to_str = {
        0: "unknown",
        1: "bottle",
        2: "bowl",
        3: "camera",
        4: "can",
        5: "laptop",
        6: "mug",
    }
    category_str_to_id = {v: k for k, v in category_id_to_str.items()}

    class Config(TypedDict, total=False):
        """Configuration dictionary for NOCSDataset.

        Attributes:
            root_dir: See NOCSDataset docstring.
            split:
                The dataset split. The following strings are supported:
                    "camera_train": 275000 images, synthetic objects + real background
                    "camera_val": 25000 images, synthetic objects + real background
                    "real_train": 4300 images in 7 scenes, real
                    "real_test": 2750 images in 6 scenes, real
            mask_pointcloud: Whether the returned pointcloud will be masked.
            normalize_pointcloud:
                Whether the returned pointcloud and position will be normalized, such
                that pointcloud centroid is at the origin.
            scale_convention:
                Which scale is returned. The following strings are supported:
                    "diagonal":
                        Length of bounding box' diagonal. This is what NOCS uses.
                    "max": Maximum side length of bounding box.
                    "half_max": Half maximum side length of bounding box.
                    "full": Bounding box side lengths. Shape (3,).
            camera_convention:
                Which camera convention is used for position and orientation. One of:
                    "opengl": x right, y up, z back
                    "opencv": x right, y down, z forward
                Note that this does not influence how the dataset is processed, only the
                returned position and quaternion.
            orientation_repr:
                Which orientation representation is used. One of:
                    "quaternion"
                    "discretized"
            orientation_grid_resolution:
                Resolution of the orientation grid.
                Only used if orientation_repr is "discretized".
            remap_y_axis:
                If not None, the NOCS y-axis will be mapped to the provided axis.
                Resulting coordinate system will always be right-handed.
                This is typically the up-axis.
                Note that NOCS object models are NOT aligned the same as ShapeNetV2.
                To get ShapeNetV2 alignment: y
                One of: "x", "y", "z", "-x", "-y", "-z"
            remap_x_axis:
                If not None, the original x-axis will be mapped to the provided axis.
                Resulting coordinate system will always be right-handed.
                Note that NOCS object models are NOT aligned the same as ShapeNetV2.
                To get ShapeNetV2 alignment: -z
                One of: "x", "y", "z", "-x", "-y", "-z"
            category_str:
                If not None, only samples from the matching category will be returned.
                See NOCSDataset.category_id_to_str for admissible category strings.
        """

        root_dir: str
        split: str
        mask_pointcloud: bool
        normalize_pointcloud: bool
        scale_convention: str
        camera_convention: str
        orientation_repr: str
        orientation_grid_resolution: int
        remap_y_axis: Optional[str]
        remap_x_axis: Optional[str]
        category_str: Optional[str]

    # TODO symmetry
    # TODO unify dataset code, adapt generated_dataset

    default_config: Config = {
        "root_dir": None,
        "split": None,
        "mask_pointcloud": False,
        "normalize_pointcloud": False,
        "camera_convention": "opengl",
        "scale_convention": "half_max",
        "orientation_repr": "quaternion",
        "orientation_grid_resolution": None,
        "category_str": None,
        "remap_y_axis": None,
        "remap_x_axis": None,
    }

    def __init__(
        self,
        config: Config,
    ) -> None:
        """Initialize the dataset.

        Args:
            config:
                Configuration dictionary of dataset. Provided dictionary will be merged
                with default_dict. See NOCSDataset.Config for supported keys.
        """
        config = yoco.load_config(config, current_dict=NOCSDataset.default_config)
        self._root_dir = config["root_dir"]
        self._split = config["split"]
        self._camera_convention = config["camera_convention"]
        self._camera = self._get_split_camera()
        self._preprocess_path = os.path.join(self._root_dir, "sdfest_pre", self._split)
        if not os.path.isdir(self._preprocess_path):
            self._preprocess_dataset()
        self._mask_pointcloud = config["mask_pointcloud"]
        self._normalize_pointcloud = config["normalize_pointcloud"]
        self._scale_convention = config["scale_convention"]
        self._sample_files = self._get_sample_files(config["category_str"])
        self._remap_y_axis = config["remap_y_axis"]
        self._remap_x_axis = config["remap_x_axis"]
        self._orientation_repr = config["orientation_repr"]
        if self._orientation_repr == "discretized":
            self._orientation_grid = so3grid.SO3Grid(
                config["orientation_grid_resolution"]
            )

    def __len__(self) -> int:
        """Return number of sample in dataset."""
        return len(self._sample_files)

    def __getitem__(self, idx: int) -> dict:
        """Return a sample of the dataset.

        Args:
            idx: Index of the instance.
        Returns:
            Sample containing the following items:
                "color"
                "depth"
                "mask"
                "pointset"
                "position"
                "orientation"
                "quaternion"
                "scale"
                "color_path"
        """
        sample_file = self._sample_files[idx]
        sample_data = pickle.load(open(sample_file, "rb"))
        sample = self._sample_from_sample_data(sample_data)
        return sample

    def _preprocess_dataset(self) -> None:
        """Create preprocessing files for the current split.

        One file per sample, which currently means per valid object mask will be
        created.

        Preprocessing will be stored on disk to {root_dir}/sdfest_pre/...
        This function will not store the preprocessing, so it still has to be loaded
        afterwards.
        """
        os.makedirs(self._preprocess_path)

        self._fix_obj_models()

        self._start_time = time.time()
        self._color_paths = self._get_color_files()

        Parallel(n_jobs=-1)(
            (
                delayed(self._preprocess_color_path)(i, color_path)
                for i, color_path in enumerate(self._color_paths)
            )
        )

        # store dictionary to map category to files
        sample_files = self._get_sample_files()
        category_str_to_files = {
            category_str: [] for category_str in NOCSDataset.category_id_to_str.values()
        }
        for sample_file in tqdm(sample_files):
            sample_data = pickle.load(open(sample_file, "rb"))
            category_id = sample_data["category_id"]
            category_str = NOCSDataset.category_id_to_str[category_id]
            _, file_name = os.path.split(sample_file)
            category_str_to_files[category_str].append(file_name)

        category_json_path = os.path.join(self._preprocess_path, "categories.json")
        with open(category_json_path, "w") as f:
            json.dump(dict(category_str_to_files), f)

        print(f"Finished preprocessing for {self._split}.", end="\033[K\n")

    def _fix_obj_models(self) -> None:
        """Fix issues with fileextensions.

        Some png files have jpg extension. This function fixes these models.
        """
        glob_pattern = os.path.join(self._root_dir, "**", "*.jpg")
        files = glob(glob_pattern, recursive=True)
        for filepath in files:
            what = imghdr.what(filepath)
            if what == "png":
                print("Fixing: ", filepath)
                folder, problematic_filename = os.path.split(filepath)
                name, _ = problematic_filename.split(".")
                fixed_filename = f"fixed_{name}.png"
                fixed_filepath = os.path.join(folder, fixed_filename)

                mtl_filepath = os.path.join(folder, "model.mtl")
                bu_mtl_filepath = os.path.join(folder, "model.mtl.old")
                copyfile(mtl_filepath, bu_mtl_filepath)

                copyfile(filepath, fixed_filepath)

    def _update_preprocess_progress(self, image_id: int) -> None:
        current_time = time.time()
        duration = current_time - self._start_time
        imgs_per_sec = image_id / duration
        if image_id > 10:
            remaining_imgs = len(self._color_paths) - image_id
            remaining_secs = remaining_imgs / imgs_per_sec
            remaining_time_str = str(datetime.timedelta(seconds=round(remaining_secs)))
        else:
            remaining_time_str = "N/A"
        print(
            f"Preprocessing image: {image_id:>10} / {len(self._color_paths)}"
            f" {image_id / len(self._color_paths) * 100:>6.2f}%"  # progress percentage
            f" Remaining time: {remaining_time_str}"  # remaining time
            "\033[K",  # clear until end of line
            end="\r",  # overwrite previous
        )

    def _preprocess_color_path(self, image_id: int, color_path: str) -> None:
        counter = 0
        self._update_preprocess_progress(image_id)

        depth_path = self._depth_path_from_color_path(color_path)
        if not os.path.isfile(depth_path):
            print(f"Missing depth file {depth_path}. Skipping.", end="\033[K\n")
            return

        mask_path = self._mask_path_from_color_path(color_path)
        meta_path = self._meta_path_from_color_path(color_path)
        meta_data = pd.read_csv(
            meta_path, sep=" ", header=None, converters={2: lambda x: str(x)}
        )
        instances_mask = self._load_mask(mask_path)
        mask_ids = np.unique(instances_mask).tolist()
        gt_id = 0  # GT only contains valid objects of interests and is 0-indexed
        for mask_id in mask_ids:

            if mask_id == 255:  # 255 is background
                continue
            match = meta_data[meta_data.iloc[:, 0] == mask_id]
            if match.empty:
                print(
                    f"Warning: mask {mask_id} not found in {meta_path}", end="\033[K\n"
                )
            elif match.shape[0] != 1:
                print(
                    f"Warning: mask {mask_id} not unique in {meta_path}", end="\033[K\n"
                )

            meta_row = match.iloc[0]
            category_id = meta_row.iloc[1]
            if category_id == 0:  # unknown / distractor object
                continue

            try:
                (
                    position,
                    orientation_q,
                    extents,
                    nocs_transform,
                ) = self._get_pose_and_scale(color_path, mask_id, gt_id, meta_row)
            except nocs_utils.PoseEstimationError:
                print(
                    "Insufficient data for pose estimation. "
                    f"Skipping {color_path}:{mask_id}.",
                    end="\033[K\n",
                )
                continue
            except ObjectError:
                print(
                    "Insufficient object mesh for pose estimation. "
                    f"Skipping {color_path}:{mask_id}.",
                    end="\033[K\n",
                )
                continue

            obj_path = self._get_obj_path(meta_row)
            sample_info = {
                "color_path": color_path,
                "depth_path": depth_path,
                "mask_path": mask_path,
                "mask_id": mask_id,
                "category_id": category_id,
                "obj_path": obj_path,
                "nocs_transform": nocs_transform,
                "position": position,
                "orientation_q": orientation_q,
                "extents": extents,
                "nocs_scale": torch.linalg.norm(extents),
                "max_extent": torch.max(extents),
            }
            out_file = os.path.join(
                self._preprocess_path, f"{image_id:08}_{counter}.pkl"
            )
            pickle.dump(sample_info, open(out_file, "wb"))
            counter += 1
            gt_id += 1

    def _get_color_files(self) -> list:
        """Return list of paths of color images of the selected split."""
        if self._split == "camera_train":
            glob_pattern = os.path.join(self._root_dir, "train", "**", "*_color.png")
            return sorted(glob(glob_pattern, recursive=True))
        elif self._split == "camera_val":
            glob_pattern = os.path.join(self._root_dir, "val", "**", "*_color.png")
            return sorted(glob(glob_pattern, recursive=True))
        elif self._split == "real_train":
            glob_pattern = os.path.join(
                self._root_dir, "real_train", "**", "*_color.png"
            )
            return sorted(glob(glob_pattern, recursive=True))
        elif self._split == "real_test":
            glob_pattern = os.path.join(
                self._root_dir, "real_test", "**", "*_color.png"
            )
            return sorted(glob(glob_pattern, recursive=True))
        else:
            raise ValueError(f"Specified split {self._split} is not supported.")

    def _get_sample_files(self, category_str: Optional[str] = None) -> list:
        """Return sorted list of sample file paths.

        Sample files are generated by NOCSDataset._preprocess_dataset.

        Args:
            category_str:
                If not None, only instances of the provided category will be returned.
        Returns:
            List of sample_data files.
        """
        glob_pattern = os.path.join(self._preprocess_path, "*.pkl")
        sample_files = glob(glob_pattern)
        sample_files.sort()
        if category_str is None:
            return sample_files
        if category_str not in NOCSDataset.category_str_to_id:
            raise ValueError(f"Unsupported category_str {category_str}.")

        category_json_path = os.path.join(self._preprocess_path, "categories.json")
        with open(category_json_path, "r") as f:
            category_str_to_filenames = json.load(f)
        filtered_sample_files = [
            os.path.join(self._preprocess_path, fn)
            for fn in category_str_to_filenames[category_str]
        ]
        return filtered_sample_files

    def _get_split_camera(self) -> None:
        """Return camera information for selected split."""
        # from: https://github.com/hughw19/NOCS_CVPR2019/blob/master/detect_eval.py
        if self._split in ["real_train", "real_test"]:
            return Camera(
                width=640,
                height=480,
                fx=591.0125,
                fy=590.16775,
                cx=322.525,
                cy=244.11084,
                pixel_center=0.0,
            )
        elif self._split in ["camera_train", "camera_val"]:
            return Camera(
                width=640,
                height=480,
                fx=577.5,
                fy=577.5,
                cx=319.5,
                cy=239.5,
                pixel_center=0.0,
            )
        else:
            raise ValueError(f"Specified split {self._split} is not supported.")

    def _sample_from_sample_data(self, sample_data: dict) -> dict:
        """Create dictionary containing a single sample."""
        color = torch.from_numpy(
            np.asarray(Image.open(sample_data["color_path"]), dtype=np.float32) / 255
        )
        depth = self._load_depth(sample_data["depth_path"])
        instances_mask = self._load_mask(sample_data["mask_path"])
        instance_mask = instances_mask == sample_data["mask_id"]

        pointcloud_mask = instance_mask if self._mask_pointcloud else None
        pointcloud = pointset_utils.depth_to_pointcloud(
            depth,
            self._camera,
            mask=pointcloud_mask,
            convention=self._camera_convention,
        )

        # adjust camera convention for position, orientation and scale
        position = pointset_utils.change_position_camera_convention(
            sample_data["position"], "opencv", self._camera_convention
        )

        # orientation / scale
        orientation_q, extents = self._change_axis_convention(
            sample_data["orientation_q"], sample_data["extents"]
        )
        orientation_q = pointset_utils.change_orientation_camera_convention(
            orientation_q, "opencv", self._camera_convention
        )
        orientation = self._quat_to_orientation_repr(orientation_q)
        scale = self._get_scale(sample_data, extents)

        # normalize pointcloud & position
        if self._normalize_pointcloud:
            pointcloud, centroid = pointset_utils.normalize_points(pointcloud)
            position = position - centroid

        sample = {
            "color": color,
            "depth": depth,
            "pointset": pointcloud,
            "mask": instance_mask,
            "position": position,
            "orientation": orientation,
            "quaternion": orientation_q,
            "scale": scale,
            "color_path": sample_data["color_path"],
            "obj_path": sample_data["obj_path"],
            "category_id": sample_data["category_id"],
            "category_str": NOCSDataset.category_id_to_str[sample_data["category_id"]],
        }
        return sample

    def _depth_path_from_color_path(self, color_path: str) -> str:
        """Return path to depth file from color filepath."""
        if self._split in ["real_train", "real_test"]:
            depth_path = color_path.replace("color", "depth")
        elif self._split in ["camera_train"]:
            depth_path = color_path.replace("color", "composed")
            depth_path = depth_path.replace("/train/", "/camera_full_depths/train/")
        elif self._split in ["camera_val"]:
            depth_path = color_path.replace("color", "composed")
            depth_path = depth_path.replace("/val/", "/camera_full_depths/val/")
        else:
            raise ValueError(f"Specified split {self._split} is not supported.")
        return depth_path

    def _mask_path_from_color_path(self, color_path: str) -> str:
        """Return path to mask file from color filepath."""
        mask_path = color_path.replace("color", "mask")
        return mask_path

    def _meta_path_from_color_path(self, color_path: str) -> str:
        """Return path to meta file from color filepath."""
        meta_path = color_path.replace("color.png", "meta.txt")
        return meta_path

    def _nocs_map_path_from_color_path(self, color_path: str) -> str:
        """Return NOCS map filepath from color filepath."""
        nocs_map_path = color_path.replace("color.png", "coord.png")
        return nocs_map_path

    def _get_pose_and_scale(
        self, color_path: str, mask_id: int, gt_id: int, meta_row: pd.Series
    ) -> tuple:
        """Return position, orientation, scale and NOCS transform.

        All of those follow OpenCV (x right, y down, z forward) convention.

        Args:
            color_path: Path to the color file.
            mask_id: Instance id in the instances mask.
            gt_id:
                Ground truth id. This is 0-indexed id of valid instances in meta file.
            meta_row:
                Matching row of meta file. Contains necessary information about mesh.

        Returns:
            position (torch.Tensor):
                Position of object center in camera frame. Shape (3,).
            quaternion (torch.Tensor):
                Orientation of object in camera frame.
                Scalar-last quaternion, shape (4,).
            extents (torch.Tensor):
                Bounding box side lengths.
            nocs_transformation (torch.Tensor):
                Transformation from centered [-0.5,0.5]^3 NOCS coordinates to camera.
        """
        gts_path = self._get_gts_path(color_path)
        obj_path = self._get_obj_path(meta_row)
        if self._split == "real_test":
            # only use gt for real test data, since there are errors in camera val
            gts_data = pickle.load(open(gts_path, "rb"))
            nocs_transform = gts_data["gt_RTs"][gt_id]
            position = nocs_transform[0:3, 3]
            rot_scale = nocs_transform[0:3, 0:3]
            nocs_scales = np.sqrt(np.sum(rot_scale ** 2, axis=0))
            rotation_matrix = rot_scale / nocs_scales[:, None]
            nocs_scale = nocs_scales[0]
        else:  # camera_train, camera_val, real_train
            # use ground truth NOCS mask to perform alignment
            (
                position,
                rotation_matrix,
                nocs_scale,
                nocs_transform,
            ) = self._estimate_object(color_path, mask_id)

        orientation_q = Rotation.from_matrix(rotation_matrix).as_quat()
        mesh_extents = self._get_mesh_extents_from_obj(obj_path)

        if "camera" in self._split:
            # CAMERA / ShapeNet meshes are normalized s.t. diagonal == 1
            # get metric extents by scaling with the diagonal
            extents = nocs_scale * mesh_extents
        elif "real" in self._split:
            # REAL object meshes are not normalized
            extents = mesh_extents
        else:
            raise ValueError(f"Specified split {self._split} is not supported.")

        position = torch.Tensor(position)
        orientation_q = torch.Tensor(orientation_q)
        extents = torch.Tensor(extents)
        nocs_transform = torch.Tensor(nocs_transform)
        return position, orientation_q, extents, nocs_transform

    def _get_gts_path(self, color_path: str) -> Optional[str]:
        """Return path to gts file from color filepath.

        Return None if split does not have ground truth information.
        """
        if self._split == "real_test":
            gts_folder = os.path.join(self._root_dir, "gts", "real_test")
        elif self._split == "camera_val":
            gts_folder = os.path.join(self._root_dir, "gts", "val")
        else:
            return None

        path = os.path.normpath(color_path)
        split_path = path.split(os.sep)
        number = path[-14:-10]
        gts_filename = f"results_{split_path[-3]}_{split_path[-2]}_{number}.pkl"
        gts_path = os.path.join(gts_folder, gts_filename)
        return gts_path

    def _get_obj_path(self, meta_row: pd.Series) -> str:
        """Return path to object file from meta data row."""
        if "camera" in self._split:  # ShapeNet mesh
            synset_id = meta_row.iloc[2]
            object_id = meta_row.iloc[3]
            obj_path = os.path.join(
                self._root_dir,
                "obj_models",
                self._split.replace("camera_", ""),
                synset_id,
                object_id,
                "model.obj",
            )
        elif "real" in self._split:  # REAL mesh
            object_id = meta_row.iloc[2]
            obj_path = os.path.join(
                self._root_dir, "obj_models", self._split, object_id + ".obj"
            )
        else:
            raise ValueError(f"Specified split {self._split} is not supported.")
        return obj_path

    def _get_mesh_extents_from_obj(self, obj_path: str) -> torch.Tensor:
        """Return maximum extent of bounding box from obj filepath.

        Note that this is normalized extent (with diagonal == 1) in the case of CAMERA
        dataset, and unnormalized (i.e., metric) extent in the case of REAL dataset.
        """
        mesh = o3d.io.read_triangle_mesh(obj_path)
        vertices = np.asarray(mesh.vertices)
        if len(vertices) == 0:
            raise ObjectError()
        extents = np.max(vertices, axis=0) - np.min(vertices, axis=0)
        return torch.Tensor(extents)

    def _load_mask(self, mask_path: str) -> torch.Tensor:
        """Load mask from mask filepath."""
        mask_img = np.asarray(Image.open(mask_path), dtype=np.uint8)
        if mask_img.ndim == 3 and mask_img.shape[2] == 4:  # CAMERA masks are RGBA
            instances_mask = mask_img[:, :, 0]  # use first channel only
        else:  # REAL masks are grayscale
            instances_mask = mask_img
        return torch.from_numpy(instances_mask)

    def _load_depth(self, depth_path: str) -> torch.Tensor:
        """Load depth from depth filepath."""
        depth = torch.from_numpy(
            np.asarray(Image.open(depth_path), dtype=np.float32) * 0.001
        )
        return depth

    def _load_nocs_map(self, nocs_map_path: str) -> torch.Tensor:
        """Load NOCS map from NOCS map filepath.

        Returns:
            NOCS map where each channel corresponds to one dimension in NOCS.
            Coordinates are normalized to [0,1], shape (H,W,3).
        """
        nocs_map = torch.from_numpy(
            np.asarray(Image.open(nocs_map_path), dtype=np.float32) / 255
        )
        # z-coordinate has to be flipped
        # see https://github.com/hughw19/NOCS_CVPR2019/blob/14dbce775c3c7c45bb7b19269bd53d68efb8f73f/dataset.py#L327 # noqa: E501
        nocs_map[:, :, 2] = 1 - nocs_map[:, :, 2]
        return nocs_map[:, :, :3]

    def _estimate_object(self, color_path: str, mask_id: int) -> tuple:
        """Estimate pose and scale through ground truth NOCS map."""
        position = rotation_matrix = scale = out_transform = None
        depth_path = self._depth_path_from_color_path(color_path)
        depth = self._load_depth(depth_path)
        mask_path = self._mask_path_from_color_path(color_path)
        instances_mask = self._load_mask(mask_path)
        instance_mask = instances_mask == mask_id
        nocs_map_path = self._nocs_map_path_from_color_path(color_path)
        nocs_map = self._load_nocs_map(nocs_map_path)
        valid_instance_mask = instance_mask * depth != 0
        nocs_map[~valid_instance_mask] = 0
        centered_nocs_points = nocs_map[valid_instance_mask] - 0.5

        measured_points = pointset_utils.depth_to_pointcloud(
            depth, self._camera, mask=valid_instance_mask, convention="opencv"
        )

        # require at least 30 point correspondences to prevent outliers
        if len(measured_points) < 30:
            raise nocs_utils.PoseEstimationError()

        # skip object if it cointains errorneous depth
        if torch.max(depth[valid_instance_mask]) > 32.0:
            print("Erroneous depth detected.", end="\033[K\n")
            raise nocs_utils.PoseEstimationError()

        (
            position,
            rotation_matrix,
            scale,
            out_transform,
        ) = nocs_utils.estimate_similarity_transform(
            centered_nocs_points, measured_points, verbose=False
        )

        if position is None:
            raise nocs_utils.PoseEstimationError()

        return position, rotation_matrix, scale, out_transform

    def _get_scale(self, sample_data: dict, extents: torch.Tensor) -> float:
        """Return scale from stored sample data and extents."""
        if self._scale_convention == "diagonal":
            return sample_data["nocs_scale"]
        elif self._scale_convention == "max":
            return sample_data["max_extent"]
        elif self._scale_convention == "half_max":
            return 0.5 * sample_data["max_extent"]
        elif self._scale_convention == "full":
            return extents
        else:
            raise ValueError(
                f"Specified scale convention {self._scale_convnetion} not supported."
            )

    def _change_axis_convention(
        self, orientation_q: torch.Tensor, extents: torch.Tensor
    ) -> tuple:
        """Adjust up-axis for orientation and extents.

        Returns:
            Tuple of position, orienation_q and extents, with specified up-axis.
        """
        if self._remap_y_axis is None and self._remap_x_axis is None:
            return orientation_q, extents
        elif self._remap_y_axis is None or self._remap_x_axis is None:
            raise ValueError("Either both or none of remap_{y,x}_axis have to be None.")

        rotation_o2n = self._get_o2n_object_rotation_matrix()
        remapped_extents = torch.abs(torch.Tensor(rotation_o2n) @ extents)

        # quaternion so far: original -> camera
        # we want a quaternion: new -> camera
        rotation_n2o = rotation_o2n.T

        quaternion_n2o = torch.from_numpy(Rotation.from_matrix(rotation_n2o).as_quat())

        remapped_orientation_q = quaternion_utils.quaternion_multiply(
            orientation_q, quaternion_n2o
        )  # new -> original -> camera

        return remapped_orientation_q, remapped_extents

    def _get_o2n_object_rotation_matrix(self) -> np.ndarray:
        """Compute rotation matrix which rotates original to new object coordinates."""
        rotation_o2n = np.zeros((3, 3))  # original to new object convention
        if self._remap_y_axis == "x":
            rotation_o2n[0, 1] = 1
        elif self._remap_y_axis == "-x":
            rotation_o2n[0, 1] = -1
        elif self._remap_y_axis == "y":
            rotation_o2n[1, 1] = 1
        elif self._remap_y_axis == "-y":
            rotation_o2n[1, 1] = -1
        elif self._remap_y_axis == "z":
            rotation_o2n[2, 1] = 1
        elif self._remap_y_axis == "-z":
            rotation_o2n[2, 1] = -1
        else:
            raise ValueError("Unsupported remap_y_axis {self.remap_y}")

        if self._remap_x_axis == "x":
            rotation_o2n[0, 0] = 1
        elif self._remap_x_axis == "-x":
            rotation_o2n[0, 0] = -1
        elif self._remap_x_axis == "y":
            rotation_o2n[1, 0] = 1
        elif self._remap_x_axis == "-y":
            rotation_o2n[1, 0] = -1
        elif self._remap_x_axis == "z":
            rotation_o2n[2, 0] = 1
        elif self._remap_x_axis == "-z":
            rotation_o2n[2, 0] = -1
        else:
            raise ValueError("Unsupported remap_x_axis {self.remap_y}")

        # infer last column
        rotation_o2n[:, 2] = 1 - np.abs(np.sum(rotation_o2n, 1))  # rows must sum to +-1
        rotation_o2n[:, 2] *= np.linalg.det(rotation_o2n)  # make special orthogonal
        if np.linalg.det(rotation_o2n) != 1.0:  # check if special orthogonal
            raise ValueError("Unsupported combination of remap_{y,x}_axis. det != 1")
        return rotation_o2n

    def _quat_to_orientation_repr(self, quaternion: torch.Tensor) -> torch.Tensor:
        """Convert quaternion to selected orientation representation.

        Args:
            quaternion:
                The quaternion to convert, scalar-last, shape (4,).
        Returns:
            The same orientation as represented by the quaternion in the chosen
            orientation representation.
        """
        if self._orientation_repr == "quaternion":
            return quaternion
        elif self._orientation_repr == "discretized":
            index = self._orientation_grid.quat_to_index(quaternion.numpy())
            return torch.tensor(
                index,
                dtype=torch.long,
            )
        else:
            raise NotImplementedError(
                f"Orientation representation {self._orientation_repr} is not supported."
            )

    def load_mesh(self, object_path: str) -> o3d.geometry.TriangleMesh:
        """Load an object mesh and adjust its object frame convention."""
        mesh = o3d.io.read_triangle_mesh(object_path)
        if self._remap_y_axis is None and self._remap_x_axis is None:
            return mesh
        elif self._remap_y_axis is None or self._remap_x_axis is None:
            raise ValueError("Either both or none of remap_{y,x}_axis have to be None.")

        rotation_o2n = self._get_o2n_object_rotation_matrix()
        mesh.rotate(
            rotation_o2n,
            center=np.array([0.0, 0.0, 0.0])[:, None],
        )
        return mesh


class ObjectError(Exception):
    """Error if something with the mesh is wrong."""

    pass
